# ══════════════════════════════════════════════════════════════════════════
# test URLs, DOIs, and ISBNs

# ──────────────────────────────────────────────────────────────────────────
# test URLs

book_url = "https://www.sciencedirect.com/book/9780123848666/an-introduction-to-dynamic-meteorology"
book_isbn = "978-0-12-384866-6"
book_doi = "https://doi.org/10.1016/C2009-0-63394-8"

#  ──────────────────────────────────────────────────────────────────────────
# DOI

doi_url = "https://doi.org/10.1002/2016JC011857"
doi_str = "10.1002/2016JC011857"
doi_bib = "@article{Houpert_2016, title={Observations of open-ocean deep convection in the northwestern Mediterranean Sea: Seasonal and interannual variability of mixing and deep water masses for the 2007-2013 Period: DEEP CONVECTION OBS. NWMED 2007-2013}, volume={121}, ISSN={2169-9275}, url={http://dx.doi.org/10.1002/2016JC011857}, DOI={10.1002/2016jc011857}, number={11}, journal={Journal of Geophysical Research: Oceans}, publisher={American Geophysical Union (AGU)}, author={Houpert, L. and Durrieu de Madron, X. and Testor, P. and Bosse, A. and D’Ortenzio, F. and Bouin, M. N. and Dausse, D. and Le Goff, H. and Kunesch, S. and Labaste, M. and Coppola, L. and Mortier, L. and Raimbault, P.}, year={2016}, month=nov, pages={8139–8171} }"

#  ──────────────────────────────────────────────────────────────────────────
# ISBN

isbn10 = "0-486-60061-0"
isbn13 = "9780486600611"
isbn_bib = """@book{9780486600611,
     title = {Fundamentals Of Astrodynamics},
    author = {Roger R. Bate and Donald D. Mueller and Jerry E. White},
      isbn = {9780486600611},
      year = {1971},
 publisher = {Courier Corporation}
}"""
